//----------------------------------------------------------------- 2013-09-29 ------------
// 電力計 無線通信版 Ver 1.0.1
//-------------------------------------------------------- written by iizuka --------------
#include <SPI.h>
#include <Mirf.h>
#include <MirfHardwareSpiDriver.h>
  
// #define DEBUG

//---------------- 環境により変更する項目 -------------------------------------------
#define FREQ    50   // 50Hz
//  #define FREQ 60   // 60Hz
#define CT_TURN                   3000  // ＣＴセンサーのコイル巻き数 CTL-10-CLS は 3000巻き
#define CH1_R_LOW_RESOLUTION       200  // 回路図のR1=200Ω  抵抗は出力が振り切れない範囲で大きい方が計測精度が良い
#define CH1_R_HIGH_RESOLUTION     3000  // 回路図のR3=3kΩ   実際の抵抗は誤差があるのでテスターで実測して
#define CH2_R_LOW_RESOLUTION       200  // 回路図のR2=200Ω  この値を変更すると精度が向上する
#define CH2_R_HIGH_RESOLUTION     3000  // 回路図のR4=3kΩ
#define SAMPLES 120                     // 1サイクル(1/50秒)に何回サンプリングするか
                                        // 大きいほど精度が高く計測できるが大きすぎると
                                        // Ａ／Ｄ変換の速度が間に合わない 150あたりが限界
#define EEP_MAX_ADDR      0x20000       // 25AA1024(1Mbit)の最大アドレス
#define EEP_NUM           4             // 実装されているEEPROMの数
#define sampling_interval 2000          // サンプリング間隔[msec] 最大65秒まで可
                                        // 但し１分以上にするとグラフ表示プログラムのデータが歯抜けになるため３０秒以下推奨
#define MAX_RECORDS       (EEP_MAX_ADDR / 4 * EEP_NUM) // EEPROMに保存できる最大計測ポイント数

//---------------- 配線を変更したら変更する項目 -------------------------------------------
// CTセンサー関連
#define PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION    3  //  Port Digital 3
#define PORT_CH1_RANGE_SELECT_LOW_RESOLUTION     2  //  Port Digital 2
#define PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION   15  //  Port Analog  1
#define PORT_CH2_RANGE_SELECT_LOW_RESOLUTION    14  //  Port Analog  0
#define PORT_ZERO_CROSS 4   // Port Digital  4
#define PORT_CH1        18  // Port Analog   4
#define PORT_CH2        16  // Port Analog   2
#define PORT_REF        17  // Port Analog   3  
#define PORT_BUZZER     19  // Port Analog   5

// EEPROM関連
#define EEPROM1_CS     8          // EEPROMのCSピン番号
#define EEPROM2_CS     7          // EEPROMのCSピン番号
#define EEPROM3_CS     6          // EEPROMのCSピン番号
#define EEPROM4_CS     5          // EEPROMのCSピン番号


//--------------- 方変換の共用体定義 -----------------------------
typedef union {
  unsigned long long_val;
  struct {
    byte LL;  // Lower address
    byte ML;
    byte MH;
    byte HH;  // Higher address
  } byte_val;    
} type_convert_long_byte; 

typedef union {
  unsigned int int_val;
  struct {
    byte L;
    byte H;
  } byte_val;    
} type_convert_int_byte; 

//---------------- 回路定数など -----------------------------------------------------------
#define SELECT          HIGH
#define NOT_SELECT      LOW

#define HIGH_RESOLUTION 0
#define LOW_RESOLUTION  1

#define REF_VOLTS    2.495              // TL431 Typ 2.495V
#define PERIOD  (1000000/FREQ/SAMPLES)  // microsec
const float pi = 3.141592;

int out_volts[SAMPLES];                 // 電流－＞電圧変換回路の出力の計測値
int ac100_sin_volts[SAMPLES];           // AC100V サインカーブ
int ref_volts_10bit;                    // TL431リファレンス電圧のA/D RAW値

//---------------- appl -----------------------------------------------------------
boolean zc_circuit_connected;

byte radio_buf_read[32];
byte radio_buf_write[32];

unsigned long watt_ch1;
unsigned long watt_ch2;
unsigned long sampling_counter;
unsigned long sampling_start_at;

char windows_datetime[20];
unsigned long seq_no_save;   // このプログラムを起動してから永久にカウントアップ
unsigned long seq_no_send;   // 送信データのシーケンシャル番号
unsigned long seq_no_sync;   // 時刻を同期させるシーケンシャル番号

bool time_request;           // 時刻定期同期のためのフラグ　trueの時に同期をかける
bool pc_ready;               // PCとの通信が確立しているときにTrueになる
bool eeprom_full;            // eepromが一杯になったときに計測をとめるためのフラグ
bool wake_up;                // 電源投入後時刻同期が行われていないときtrue
bool ping;                   // trueの時にpingを実行する       
      
//------------------------------------------------------------------------------------------
// SEQ_No からEEPROMのアドレスに変換する
//------------------------------------------------------------------------------------------
unsigned long seq_to_addr( long seq_no )
{
  return ( seq_no % ((EEP_MAX_ADDR * EEP_NUM) >> 2) ) << 2;
}

//------------------------------------------------------------------------------------------
// EEPROM 25AA1024　に書き込む　
//    実装されている24AA1025の数を判断している
//    複数EEPROMのアドレスを連続アドレスとして扱う
//
// 制約：EEPROMの256バイトページ境界を判断していないのでこれをまたぐ書き込みは不可
// RETURN: 次に書き込むべきアドレス
//------------------------------------------------------------------------------------------
void spi_eeprom_write( unsigned long addr, byte data[], byte len )
{
  byte index;
  type_convert_long_byte eep_addr;
  byte chip_number;
  
  chip_number = (addr / EEP_MAX_ADDR ) + 1;
  spi_eeprom_chipselect_LOW( chip_number );  // eeprom chip select enable 
  SPI.transfer( B00000110 );                 // eeprom enable write operation Instruction
  spi_eeprom_chipselect_HIGH( chip_number ); // eeprom chip select disable     
  
  spi_eeprom_chipselect_LOW( chip_number );  // eeprom chip select enable 
  SPI.transfer( B00000010 );                 // eeprom Write Instruction
  eep_addr.long_val = addr;
  SPI.transfer( eep_addr.byte_val.MH );            
  SPI.transfer( eep_addr.byte_val.ML );
  SPI.transfer( eep_addr.byte_val.LL );
  for( index = 0 ; index < len ; index++ ) {
    SPI.transfer( data[index]);
  }
  spi_eeprom_chipselect_HIGH( chip_number ); // eeprom chip select disable
  delay(6);
}

//------------------------------------------------------------------------------------------
// EEPROM 25AA1024　から読み出す
//    実装されている24AA1025の数を判断している
//    複数EEPROMのアドレスを連続アドレスとして扱う
//
// 制約：EEPROMの256バイトページ境界を判断していないのでこれをまたぐ読み出しは不可
//------------------------------------------------------------------------------------------
void spi_eeprom_read( unsigned long addr, byte data[], byte len )
{
  byte index;
  type_convert_long_byte eep_addr;
  byte chip_number;
  
  chip_number = ( addr / EEP_MAX_ADDR ) + 1;
  
  spi_eeprom_chipselect_LOW( chip_number );  // eeprom chip select enable 
  SPI.transfer( B00000011 );                 // eeprom Read Instruction
  eep_addr.long_val = addr;
  SPI.transfer( eep_addr.byte_val.MH );            
  SPI.transfer( eep_addr.byte_val.ML );
  SPI.transfer( eep_addr.byte_val.LL );
  for( index = 0 ; index < len ; index++ ) {
    data[index] = SPI.transfer(0);
  }
  spi_eeprom_chipselect_HIGH( chip_number ); // eeprom chip select disable
}

//------------------------------------------------------------------
// EEPROM disable
//------------------------------------------------------------------
void spi_eeprom_chipselect_HIGH( byte chip_no )
{
  switch( chip_no ) {
    case 1: digitalWrite( EEPROM1_CS, HIGH ); break;
    case 2: digitalWrite( EEPROM2_CS, HIGH ); break;
    case 3: digitalWrite( EEPROM3_CS, HIGH ); break;
    case 4: digitalWrite( EEPROM4_CS, HIGH ); break;
  }
}  
    
//------------------------------------------------------------------
// EEPROM enable
//------------------------------------------------------------------
void spi_eeprom_chipselect_LOW( byte chip_no )
{
  switch( chip_no ) {
    case 1: digitalWrite( EEPROM1_CS, LOW ); break;
    case 2: digitalWrite( EEPROM2_CS, LOW ); break;
    case 3: digitalWrite( EEPROM3_CS, LOW ); break;
    case 4: digitalWrite( EEPROM4_CS, LOW ); break;
  }
}    

//------------------------------------------------------------------------------------------
// ＡＣ１００Ｖサイン波形の生成 
// １／４サイクルでもよいがメモリー不足ではないので１サイクル分生成している
//------------------------------------------------------------------------------------------
void generate_ac100v_sine_curve()
{
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    ac100_sin_volts[time_index] = abs(sin( ( 2 * pi ) / SAMPLES * time_index )) * sqrt(2) * 100; 
  }
}  

//------------------------------------------------------------------------------------------
// １０ビットの計測値を電流値[mA]に換算する
//
// INPUUT:
//   volt     -> 計測対象機器の10ビット計測値
//   ref_volt -> TL431のA/D計測値
//   R_range  -> 計測レンジ
//------------------------------------------------------------------------------------------
float convert_10bit_to_milli_amp( int volt, int ref_volt, int R_range )
{
  return ((float)volt * REF_VOLTS * CT_TURN) / ((float)ref_volt * (float)R_range);
} 

//------------------------------------------------------------------------------------------
// １サイクルぶんをサンプリングする
//
// INPUT:
//   port  -> arduinoアナログポート番号
//   range -> HIGH_RESOLUSION / LOW_RESOLUTION
//------------------------------------------------------------------------------------------
void ad_get( int port, int range )
{
  long time_next_period;
  long period = PERIOD;
 
  if( range == HIGH_RESOLUTION ) {
    if( port == PORT_CH1 ) {
      digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
      digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );
    } else {  // PORT_CH2
      digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
      digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );
    }
  } else {  // LOW_RESOLUTION
    if( port == PORT_CH1 ) {
      digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
      digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
    } else {  // PORT_CH2
      digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
      digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
    }
  }  
  delay(1); // デジタル回路が動くと電圧が変動したりするので1ms程度待っている。たぶん効果はほとんど無い。気分の問題。

  while(digitalRead(PORT_ZERO_CROSS) == LOW ) {}  // AC100Vの電圧がゼロになるまで待つ
  time_next_period = micros();                    //  プルアップされているのでAC100につながなくてもループを抜ける
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    out_volts[time_index] = analogRead( port );
    time_next_period += period;
    while( micros() < time_next_period ) {
//      digitalWrite( PORT_LED, HIGH );  // LEDが点灯しないならA/D変換速度が間に合っていないので
                                         // サンプル数を小さくすること
    }
  }  
}

//------------------------------------------------------------------------------------------
// リファレンス電圧のＡ／Ｄ変換１０ビット値を得る
// １回Ａ／Ｄ変換すれば済むが、Ａ／Ｄの揺らぎを考慮し１０回取得して平均値を求めている
// INPUT:
//   port  -> arduinoアナログポート番号 = TL431出力
//------------------------------------------------------------------------------------------
float get_ref_volt( int port )
{
  int total = 0;
  
  for( int t = 0 ; t < 10 ; t++ ) {
    delay(1);
    total += analogRead( port );
  }  

  return (float)total / 10;
}

//------------------------------------------------------------------------------------------
// 計測値のオフセット値を求める
// 機器の電流波形は正負均等に触れると見なし、平均値をオフセットとする 値は 大体 512 になる
//------------------------------------------------------------------------------------------
float calc_offset()
{
  long total = 0;

  for(int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {  
    total += (long)out_volts[time_index];
  }

  return (float)total / SAMPLES;
}

//------------------------------------------------------------------------------------------
// １ビットあたりの電流値[mA]を求める
//
// INPUT:
//    range -> HIGH_RESOLUSION / LOW_RESOLUTION
//------------------------------------------------------------------------------------------
float get_amp_div( int port, int range )
{
  if( range == HIGH_RESOLUTION ) {
    if( port = PORT_CH1 ) {
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH1_R_HIGH_RESOLUTION) * 1000;
    } else { // CH2
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH2_R_HIGH_RESOLUTION) * 1000;
    }
  } else { // LOW_RESOLUTION
    if( port = PORT_CH1 ) {
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH1_R_LOW_RESOLUTION) * 1000;
    } else { // CH2
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH2_R_LOW_RESOLUTION) * 1000;
    }
  }
}

//------------------------------------------------------------------------------------------
// ブザーを０．２秒鳴らす
//------------------------------------------------------------------------------------------  
void beep()
{
  for( int i = 0 ; i < 100 ; i++ ) {
    digitalWrite( PORT_BUZZER, HIGH );
    delay(1);
    digitalWrite( PORT_BUZZER, LOW  );
    delay(1);
  }    
}

//------------------------------------------------------------------------------------------
// ブザーを0.04秒鳴らす
//------------------------------------------------------------------------------------------  
void beep_short()
{
  for( int i = 0 ; i < 20 ; i++ ) {
    digitalWrite( PORT_BUZZER, HIGH );
    delay(1);
    digitalWrite( PORT_BUZZER, LOW  );
    delay(1);
  }    
}

//------------------------------------------------------------------------------------------ 
// 消費電力の計算 ゼロクロス点からサンプリングされた波形であることを前提として電力を計算する
//------------------------------------------------------------------------------------------  
unsigned long calc_watt_zc( int port, int range )
{
  float offset;
  float total;
  float temp_volt;
  float amp_div;
  
  offset = calc_offset();

  amp_div = get_amp_div( port, range );
  total = 0.0;
  for( int time_index ; time_index < SAMPLES ; time_index++ ) {  // １サイクルぶん電力を積算
    temp_volt = abs((float)out_volts[time_index] - offset);
    total += (temp_volt * amp_div * (float)ac100_sin_volts[ time_index ] / 1000);
  }  

  return (unsigned int)(total / SAMPLES);
}

//------------------------------------------------------------------------------------------
// 電流波形がSIN波になる場合の電流実効値[mA]を求める
// AC100Vを接続していないときはこの計算式を使う。力率１００％なら誤差は無い
// 平均値に波形率1.11を掛けても良いが定義通り 計算式 √avg(Σ計測値^2) で計算している
//
// INPUT:
//   out_volts[SAMPLES]   extern
//   port  -> arduino analog port
//   range -> HIGH_RESOLUSION / LOW_RESOLUTION
//------------------------------------------------------------------------------------------
float calc_rms_watt( int port, int range )
{
  float total;
  float temp_volt;
  float amp_div;
  float offset;
  
  offset = calc_offset();
  total = 0.0;
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    temp_volt = abs((float)out_volts[time_index] - offset);
    total += temp_volt * temp_volt;
  }
  amp_div = get_amp_div( port, range );
  
  return sqrt( total / SAMPLES ) * amp_div * 100 / 1000;  
}

//------------------------------------------------------------------------------------------
// タイムアウト付き無線受信
//
// timeout: タイムアウト時間[msec]
//------------------------------------------------------------------------------------------
bool radio_recieve( byte buf[], unsigned long timeout )
{
  unsigned long millis_tmp;
  bool millis_overflow = false;
  
  unsigned long started_waiting_at = millis();
  while(!Mirf.dataReady()){
    if( millis_diff(started_waiting_at) > timeout ) {
      return false;
    }
  }    
  Mirf.getData( buf );
  return true;
}

//------------------------------------------------------------------------------------------
// 無線送信
//------------------------------------------------------------------------------------------
void radio_write( byte buf[] )
{
  Mirf.send( buf );
  
  while( Mirf.isSending() ){}
}

//------------------------------------------------------------------------------------------
// 引数と現在値とのミリ秒差を返す　
//   millis()の50日オーバーフローは回避されるが時間差が3.1日(=0x10000000msec)を越えると誤動作する
//------------------------------------------------------------------------------------------
unsigned long millis_diff(unsigned long millis_from)
{
  if( millis_from > 0xF0000000 ) {
    millis_from += 0x10000000;   // 強制的にオーバーフローさせる
    return (millis() + 0x10000000) - millis_from;
  } else {
    return millis() - millis_from;
  }
}

//------------------------------------------------------------------------------------------
// AC100Vゼロクロス検出回路が接続されているか確認
//------------------------------------------------------------------------------------------
boolean zc_circuit_connect()
{
  unsigned long ms = millis();
    
  while( (millis() - ms) < 200 ) {
    if( digitalRead(PORT_ZERO_CROSS) == LOW ) {
      return true;
    }
  }
  
  beep();
  delay(500);
  beep();
  delay(500);
  beep();
  delay(500);
  
  return false;
}

//------------------------------------------------------------------------------------------
// EEPROMテスト
//------------------------------------------------------------------------------------------
void eeprom_test()
{
  byte write_data[1];
  byte read_data[1];
  unsigned long addr;
  
  write_data[0] = 0xAA;
  for( byte i = 0 ; i < EEP_NUM ; i++ ) {
    addr = seq_to_addr( 32768 * i );
    spi_eeprom_write( addr, write_data, 1 );
    spi_eeprom_read(  addr, read_data,  1 );
    if( read_data[0] != 0xAA ) {
      while( true ) {
        delay(2500);
        beep();
      }
    } 
  }
}

//------------------------------------------------------------------------------------------
// arduino setup
//------------------------------------------------------------------------------------------  
void setup()
{
  pinMode(PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, OUTPUT);
  pinMode(PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  OUTPUT);
  pinMode(PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, OUTPUT);
  pinMode(PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  OUTPUT);

  pinMode(PORT_BUZZER, OUTPUT);
  
  digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
  digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
  digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
  digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );

  pinMode(EEPROM1_CS, OUTPUT);
  pinMode(EEPROM2_CS, OUTPUT);
  pinMode(EEPROM3_CS, OUTPUT);
  pinMode(EEPROM4_CS, OUTPUT);
  digitalWrite( EEPROM1_CS, HIGH );
  digitalWrite( EEPROM2_CS, HIGH );
  digitalWrite( EEPROM3_CS, HIGH );
  digitalWrite( EEPROM4_CS, HIGH );
  
  SPI.begin();
  eeprom_test();
  
  // 通信開始
  Mirf.csnPin = 10;
  Mirf.cePin  = 9;
  Mirf.spi = &MirfHardwareSpi;
  Mirf.init();

  Mirf.setRADDR((byte *)"serv2");
  Mirf.setTADDR((byte *)"clie2");
  Mirf.payload = 32;
  Mirf.channel = 3;
  Mirf.config();
  
  zc_circuit_connected = zc_circuit_connect();
  
  generate_ac100v_sine_curve();
  ref_volts_10bit = get_ref_volt( PORT_REF );
  
#ifdef DEBUG  
  Serial.begin(57600);
  Serial.println("start client");
#endif

  beep();
  delay(1000);
    
  time_request = true;         
  pc_ready     = false;          
  eeprom_full  = false;   
  wake_up      = true;
  ping         = false;

  seq_no_sync = 0xFFFFFFFF;

  seq_no_save = 0; // 次に保存すべきSEQ_No
  seq_no_send = 0; // 次に送信すべきSEQ_No
  watt_ch1 = 0;
  watt_ch2 = 0;
  sampling_counter = 0;
  
  sampling_start_at = millis();
}

//------------------------------------------------------------------------------------------
// arduino loop
//------------------------------------------------------------------------------------------  
void loop()
{
  type_convert_int_byte   val;
  type_convert_long_byte  v;
  byte eep_buf[4];
 
  unsigned long eep_send_addr;
  unsigned long eep_save_addr;
  byte send_samples = 0;
  
  //---------- A/D convert ----------------------------------------
  if( ! eeprom_full ) {
    if( zc_circuit_connected ) {
      ad_get( PORT_CH1, LOW_RESOLUTION );
      watt_ch1 += calc_watt_zc( PORT_CH1, LOW_RESOLUTION ); 
      ad_get( PORT_CH2, LOW_RESOLUTION );
      watt_ch2 += calc_watt_zc( PORT_CH2, LOW_RESOLUTION );
    } else {
      ad_get( PORT_CH1, LOW_RESOLUTION );
      watt_ch1 += calc_rms_watt( PORT_CH1, LOW_RESOLUTION ); 
      ad_get( PORT_CH2, LOW_RESOLUTION );
      watt_ch2 += calc_rms_watt( PORT_CH2, LOW_RESOLUTION );
    }     
    sampling_counter++;
  }

  //---------- store EEPROM ---------------------------------------
  if( millis_diff( sampling_start_at ) > sampling_interval ) {
    if( ! eeprom_full ) {
      val.int_val = (unsigned int)( watt_ch1 / sampling_counter );
      eep_buf[0] = val.byte_val.L;
      eep_buf[1] = val.byte_val.H;
      val.int_val = (unsigned int)( watt_ch2 / sampling_counter );
      eep_buf[2] = val.byte_val.L;
      eep_buf[3] = val.byte_val.H;
      eep_save_addr = seq_to_addr( seq_no_save );
      spi_eeprom_write( eep_save_addr, eep_buf, 4 );
      if( seq_no_send + MAX_RECORDS < seq_no_save  ) {
        eeprom_full = true;
      }
      watt_ch1 = 0;
      watt_ch2 = 0;
      sampling_counter = 0;
    }
    #ifdef DEBUG  
      Serial.print("save ");
      Serial.println(seq_no_save);
    #endif
    seq_no_save++;
    if( (seq_no_save % 10) == 0 ) {
      time_request = true;
    }
    sampling_start_at = sampling_start_at + sampling_interval;
    if( wake_up ) {
      seq_no_save = 0;
      time_request = true;
    }
    ping = true;
  }

  //--------- Ping ------------------------------------
  if( !pc_ready && ping) {
    ping = false;
    radio_buf_write[0] = 'P';
    radio_write( radio_buf_write );
  }

  //---------- PCと接続が確立されているときの処理 --------
  if( pc_ready ) {
    //--------- 時刻を要求する ---------------------------
    if( time_request ) {
      time_request = false;
      radio_buf_write[0] = 'T';
      radio_write( radio_buf_write );
    }

    //--------- 電力計測値を送信 -------------------------  
    if( seq_no_send < seq_no_save ) {
      byte send_samples = 0;
      radio_buf_write[0] = 'D';
      v.long_val = seq_no_send;
      radio_buf_write[3] = v.byte_val.LL;
      radio_buf_write[4] = v.byte_val.ML;
      radio_buf_write[5] = v.byte_val.MH;
      radio_buf_write[6] = v.byte_val.HH;
      
      if( seq_no_send == seq_no_sync ) {
        radio_buf_write[1] = 1;
        radio_buf_write[2] = 'T';
        eep_send_addr = seq_to_addr(seq_no_send);
        spi_eeprom_read( eep_send_addr, &radio_buf_write[7], 4 );
        for( int index = 0 ; index < 19 ; index++ ) {
          radio_buf_write[11+index] = windows_datetime[index];
        }
        send_samples = 1;
      } else {
        radio_buf_write[2] = ' ';
        send_samples = 0;
        for( byte count = 0 ; count < 6 ; count++ ) {
          eep_send_addr = seq_to_addr(seq_no_send);
          spi_eeprom_read( eep_send_addr, &radio_buf_write[ 7 + count * 4 ], 4 );
          radio_buf_write[1] = count + 1;
          send_samples++;
          if( (seq_no_send == seq_no_sync) || ((seq_no_send + send_samples)== seq_no_save) ) {
            break;
          }
        }
      }
    #ifdef DEBUG  
      Serial.print("send ");
      Serial.println(seq_no_send);
    #endif
      radio_write( radio_buf_write );
      seq_no_send += send_samples;
        /*
        #ifdef DEBUG  
          Serial.print("send records ");
          Serial.println(send_samples);
        #endif
        #ifdef DEBUG  
          Serial.print("send ");
          Serial.println(seq_no_send);
        #endif
        seq_no_send += send_samples;
      } else {
        #ifdef DEBUG 
          Serial.println("err  send D records");
        #endif
      }
      */
    }  
  } 
  
  //---------- PCからの応答/指令を受信 ---------------------------
  if( radio_recieve( radio_buf_read, 10 ) ) {
    switch( radio_buf_read[0] ) {
      case 't':
        for( int i = 0 ; i < 19 ; i++ ) {
          windows_datetime[i] = radio_buf_read[i+1];
        }
        seq_no_sync = seq_no_save;
        wake_up = false;
        break;
      case 'W':
        pc_ready = false;
        break;
      case 'p':
        switch( radio_buf_read[1] ) {
          case 'R':
            pc_ready = true;
            break;
          case 'N':
            pc_ready = false;
            break;
        }
        break;
      case 'd':
        beep_short();
        unsigned long seq_diff;
        v.byte_val.LL = radio_buf_read[1];
        v.byte_val.ML = radio_buf_read[2]; 
        v.byte_val.MH = radio_buf_read[3]; 
        v.byte_val.HH = radio_buf_read[4];
 
    #ifdef DEBUG  
      Serial.println("rew ");
      Serial.println(v.long_val);
    #endif     
        unsigned long start_record_no; 
        if( seq_no_save < MAX_RECORDS ) {
          start_record_no = 0;
        } else {
          start_record_no = seq_no_save - MAX_RECORDS;
        }          
        if( start_record_no < v.long_val && v.long_val < seq_no_save ) { // EEPROMメモリー領域外の要求を受け付けない
          seq_no_send = v.long_val;
          delay(300);
          beep;
          delay(300);
          beep;
          delay(300);
        }       
        break;
    }
  }
}

//------------------------------------------------------------------ E N D ------------------

